defmodule ClickHouse.Query do
  @moduledoc """
  A ClickHouse query.
  """

  alias ClickHouse.DataType

  @values_regex ~r/\bVALUES|values\b/i
  @format_regex ~r/(?<= (?:FORMAT|format) )(\w+)/

  defstruct [
    :format,
    :command,
    :statement
  ]

  @typedoc """
  The type of command being exectued in the query.
  """
  @type command :: :create | :select | :insert | :alter | :unknown

  @typedoc """
  A ClickHouse query
  """
  @type t :: %__MODULE__{
          format: ClickHouse.Format.t() | nil,
          command: command(),
          statement: iolist()
        }

  ################################
  # Public API
  ################################

  @doc """
  Prepares a new ClickHouse query.
  """
  @spec prepare(ClickHouse.Client.t(), ClickHouse.statement(), ClickHouse.params()) ::
          ClickHouse.Query.t()
  def prepare(client, statement, params \\ []) do
    statement = strip_newlines(statement)

    %__MODULE__{}
    |> put_command(statement)
    |> put_format(statement, client.formats)
    |> encode(statement, params)
  end

  ################################
  # Private API
  ################################

  defp strip_newlines(statement) do
    statement
    |> String.replace("\n", " ")
    |> String.trim_leading()
    |> String.trim_trailing()
  end

  defp put_command(query, statement) do
    command = parse_command(statement)
    %{query | command: command}
  end

  defp parse_command("SELECT" <> _), do: :select
  defp parse_command("select" <> _), do: :select
  defp parse_command("INSERT" <> _), do: :insert
  defp parse_command("insert" <> _), do: :insert
  defp parse_command("CREATE" <> _), do: :create
  defp parse_command("create" <> _), do: :create
  defp parse_command("ALTER" <> _), do: :alter
  defp parse_command("alter" <> _), do: :alter
  defp parse_command(_), do: :unknown

  defp put_format(%{command: :insert} = query, statement, formats) do
    if Regex.match?(@values_regex, statement) do
      %{query | format: ClickHouse.Format.Values}
    else
      %{query | format: parse_format(statement, formats)}
    end
  end

  defp put_format(query, statement, formats) do
    %{query | format: parse_format(statement, formats)}
  end

  defp parse_format(statement, formats) do
    case Regex.run(@format_regex, statement, capture: :first) do
      [format] -> fetch_format!(formats, format)
      _ -> nil
    end
  end

  defp fetch_format!(formats, format) do
    case Map.fetch(formats, format) do
      {:ok, format} ->
        format

      _ ->
        raise ArgumentError, """
        Format '#{format}' not implemented. \
        Please see ClickHouse.Format documentation for more information.
        """
    end
  end

  defp encode(%__MODULE__{command: :insert} = query, statement, [rows]) do
    statement = String.trim_trailing(statement, "?")
    params = query.format.encode(rows)
    %{query | statement: [statement, params]}
  end

  defp encode(%__MODULE__{command: :insert}, _statement, [_, _ | _]) do
    raise ArgumentError, """
    Insert query must contain a single parameter representing all rows.
    """
  end

  defp encode(%__MODULE__{command: command} = query, statement, params)
       when command in [:select, :create, :alter, :unknown] do
    statement = String.split(statement, "")
    statement = do_encode(statement, params)
    %{query | statement: statement}
  end

  defp do_encode(parts, params, acc \\ [])

  defp do_encode([], _, acc), do: acc

  defp do_encode(["?" | _], [], _acc) do
    raise ArgumentError, """
    Query parameters do not match the number of embedded '?'.
    """
  end

  defp do_encode(["?"], [data], acc) do
    [acc | [DataType.encode(data)]]
  end

  defp do_encode(["?" | parts], [data | params], acc) do
    do_encode(parts, params, [acc | [DataType.encode(data)]])
  end

  defp do_encode([part], _params, acc) do
    [acc | [part]]
  end

  defp do_encode([part | parts], params, acc) do
    do_encode(parts, params, [acc | [part]])
  end

  defimpl String.Chars do
    def to_string(%{statement: statement}) do
      IO.iodata_to_binary(statement)
    end
  end

  defimpl Inspect do
    def inspect(%{statement: statement}, _) do
      "#ClickHouse.Query<\"#{IO.iodata_to_binary(statement)}\">"
    end
  end
end
